/* packetDump2.c
 *
 * Copyright (C) 2004 Oki Telecommunication Systems Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "packetDump.h"
#include "wtap.h"
#include "wtap-int.h"

#ifndef	MAX_PATH
#define	MAX_PATH	260
#endif	//MAX_PATH

//oCgI[_[ϊ}N
#define ntohs(a)	(((a & 0xff00) >> 8) | ((a & 0x00ff) << 8))
#define ntohl(a)	(((a & 0xff000000) >> 24) | ((a & 0x00ff0000) >> 8) | ((a & 0x0000ff00) << 8) | ((a & 0x000000ff) << 24))
#define htons(a)	(((a & 0xff00) >> 8) | ((a & 0x00ff) << 8))
#define htonl(a)	(((a & 0xff000000) >> 24) | ((a & 0x00ff0000) >> 8) | ((a & 0x0000ff00) << 8) | ((a & 0x000000ff) << 24))

//Ethernett[wb_
typedef struct _ether_hdr {
	guint8	dst[6];		//Ml`bAhX
	guint8	src[6];		//Ml`bAhX
	guint16	frametype;	//t[^Cv
} ether_hdr;
#define FRAMETYPE_IP	0x0800

//howb_
typedef struct _ip_hdr {
    guint8  ip_v_hl; /* combines ip_v and ip_hl */
    guint8  ip_tos;
    guint16 ip_len;
    guint16 ip_id;
    guint16 ip_off;
    guint8  ip_ttl;
    guint8  ip_p;		//protocol(TCP, UDP, ...)
    guint16 ip_sum;
	union {
	    guint8 ip_src[4];
		guint32 ip_src4;
	};
	union {
	    guint8 ip_dst[4];
		guint32 ip_dst4;
	};
} ip_hdr;
#define IPv4	(0x04)

//sbowb_
typedef struct _tcp_hdr {
	guint16 th_sport;
	guint16 th_dport;
	guint32 th_seq;
	guint32 th_ack;
	guint8  th_hlen;
	guint8  th_flags;
	guint16 th_window;
	guint16 th_checksum;
	guint16 th_urgentptr;
} tcp_hdr;

//tcowb_
typedef struct _udp_hdr {
	guint16 uh_sport;
	guint16 uh_dport;
	guint16 uh_ulen;
	guint16 uh_sum;
} udp_hdr;

//pPbg\
typedef struct _packet_info {
	guint32		src_ip;
	guint32		dst_ip;
	guint16		src_port;
	guint16		dst_port;
	guint16		protocol;
	guint16		reserve;
	guint8*		buffer;
	guint32		size;
	guint32		realsize;
} packet_info;

int dump_target(wtap_dumper *wdh, packet_info* pinfo, int *err)
{
	char			outputfile[MAX_PATH];
	char*			protocol_name;
	FILE*			fh;
	int				i;
	int				rc;
	union {
		guint32		ip4;
		guint8		ip[4];
	}u1;
	union {
		guint32		ip4;
		guint8		ip[4];
	}u2;
	struct _T_PACKET_DUMP_LIST**	pptPacketDumpList;

	if(wdh->fh->lOutputCountFlag == 1){
		wdh->fh->lOutputFileCount++;
		return(ERROR_NONE);
	}

	for(pptPacketDumpList = &wdh->fh->ptPacketDumpList;
		*pptPacketDumpList != NULL;
		pptPacketDumpList = &(*pptPacketDumpList)->ptPacketDumpListNext)
	{
	}
	*pptPacketDumpList = malloc(sizeof(**pptPacketDumpList));
	if(*pptPacketDumpList == NULL)
	{
		return(ERROR_CAN_NOT_ALLOCATE_MEMORY);
	}
	memset(*pptPacketDumpList, 0, sizeof(**pptPacketDumpList));
	rc = sprintf(
		outputfile, "%s%s-%04d.dat",
		wdh->fh->cOutputFoldername,
		wdh->fh->cBaseFilename,
		wdh->fh->lOutputFileCount + 1);
	if(rc < 0 || MAX_PATH <= rc)
	{
		*err = errno;
		return(FALSE);
	}
	strcpy((*pptPacketDumpList)->cFilename, outputfile);
	fh = fopen(outputfile, "wb");
	if(fh == NULL)
	{
		*err = errno;
		return(FALSE);
	}
	switch(pinfo->protocol)
	{
	case PROTOCOL_TCP:
		protocol_name = "TCP";
		break;
	case PROTOCOL_UDP:
		protocol_name = "UDP";
		break;
	default:
		protocol_name = "";
	}
	u1.ip4 = htonl(pinfo->src_ip);
	u2.ip4 = htonl(pinfo->dst_ip);
	i = sprintf(
		(*pptPacketDumpList)->cComment,
		";%s %d.%d.%d.%d(%d) -> %d.%d.%d.%d(%d) size = %d / %d bytes",
		protocol_name,
		u1.ip[0], u1.ip[1], u1.ip[2], u1.ip[3], pinfo->src_port,
		u2.ip[0], u2.ip[1], u2.ip[2], u2.ip[3], pinfo->dst_port,
		pinfo->size, pinfo->realsize
	);
	if(i <= 0 || sizeof((*pptPacketDumpList)->cComment) <= i)
	{
		fclose(fh);
		return(ERROR_DUMP_FAILED);
	}
	rc = fprintf(fh, (*pptPacketDumpList)->cComment);
	if(rc != i)
	{
		if(rc == 0 && ferror(fh))
			*err = errno;
		else
			*err = WTAP_ERR_SHORT_WRITE;
		fclose(fh);
		return(ERROR_DUMP_FAILED);
	}
	for(i = 0; (unsigned int)i < pinfo->size; i++)
	{
		if(i % 16 == 0)
		{
			rc = fprintf(fh, "\n");
			if(rc <= 0)
			{
				if(rc == 0 && ferror(fh))
					*err = errno;
				else
					*err = WTAP_ERR_SHORT_WRITE;
				fclose(fh);
				return(ERROR_DUMP_FAILED);
			}
		}
		rc = fprintf(fh, "%02X ", pinfo->buffer[i]);
		if(rc < 3)
		{
			if(rc == 0 && ferror(fh))
				*err = errno;
			else
				*err = WTAP_ERR_SHORT_WRITE;
			fclose(fh);
			return(ERROR_DUMP_FAILED);
		}
	}
	fprintf(fh, "\n");
// 2004/07/02 change start
//	fclose(fh);
	rc = fclose(fh);
	if(rc != 0)
	{
		*err = errno;
		return(ERROR_DUMP_FAILED);
	}
// 2004/07/02 change end
	if(pinfo->size > 16)
	{
		rc = 16;
	}
	else
	{
		rc = pinfo->size;
	}
	for(i = 0; i < rc ; i++)
	{
		sprintf(&(*pptPacketDumpList)->cData[i * 3], "%02X ", pinfo->buffer[i]);
	}
	wdh->fh->lOutputFileCount++;
	return(ERROR_NONE);
}

int checkTCP(wtap_dumper *wdh, packet_info* pinfo, int* err)
{
	tcp_hdr	tcphdr;
	int		offset;
	int		rc;

	if(pinfo->size < sizeof(tcp_hdr))
	{
		return(ERROR_PACKET_TOO_SHORT);
	}
	tcphdr = *(tcp_hdr*)pinfo->buffer;
	tcphdr.th_sport		= ntohs(tcphdr.th_sport);
	tcphdr.th_dport		= ntohs(tcphdr.th_dport);
	tcphdr.th_seq		= ntohl(tcphdr.th_seq);
	tcphdr.th_ack		= ntohl(tcphdr.th_ack);
	tcphdr.th_hlen		= tcphdr.th_hlen & 0xf0;
	tcphdr.th_flags		= tcphdr.th_flags & 0x3f;
	tcphdr.th_window	= ntohs(tcphdr.th_window);
	tcphdr.th_checksum	= ntohs(tcphdr.th_checksum);
	tcphdr.th_urgentptr	= ntohs(tcphdr.th_urgentptr);

	if(wdh->fh->usSourcePort != 0)
	{
		if(wdh->fh->usSourcePort != tcphdr.th_sport)
		{
			return(ERROR_PACKET_NOT_TARGET);
		}
	}
	if(wdh->fh->usDestinationPort != 0)
	{
		if(wdh->fh->usDestinationPort != tcphdr.th_dport)
		{
			return(ERROR_PACKET_NOT_TARGET);
		}
	}

	offset = (tcphdr.th_hlen >> 4) << 2;
	if(pinfo->size < (unsigned int)offset)
	{
		return(ERROR_PACKET_TOO_SHORT);
	}
	if(pinfo->size == (unsigned int)offset)
	{
		return(ERROR_PACKET_NO_DATA);
	}
	pinfo->src_port	= tcphdr.th_sport;
	pinfo->dst_port	= tcphdr.th_dport;
	pinfo->size		= pinfo->size - offset;
	pinfo->realsize	= pinfo->realsize - offset;
	pinfo->buffer	= pinfo->buffer + offset;
	pinfo->protocol	= PROTOCOL_TCP;
	rc = dump_target(wdh, pinfo, err);
	return(rc);
}

int checkUDP(wtap_dumper *wdh, packet_info* pinfo, int* err)
{
	udp_hdr	udphdr;
	int		offset;
	int		rc;

	if(pinfo->size < sizeof(udp_hdr))
	{
		return(ERROR_PACKET_TOO_SHORT);
	}
	if(pinfo->size == sizeof(udp_hdr))
	{
		return(ERROR_PACKET_NO_DATA);
	}

	udphdr = *(udp_hdr*)pinfo->buffer;
	udphdr.uh_sport	= ntohs(udphdr.uh_sport);
	udphdr.uh_dport	= ntohs(udphdr.uh_dport);
	udphdr.uh_ulen	= ntohs(udphdr.uh_ulen);
	udphdr.uh_sum	= ntohs(udphdr.uh_sum);

	if(wdh->fh->usSourcePort != 0)
	{
		if(wdh->fh->usSourcePort != udphdr.uh_sport)
		{
			return(ERROR_PACKET_NOT_TARGET);
		}
	}
	if(wdh->fh->usDestinationPort != 0)
	{
		if(wdh->fh->usDestinationPort != udphdr.uh_dport)
		{
			return(ERROR_PACKET_NOT_TARGET);
		}
	}

	offset = sizeof(udp_hdr);
	pinfo->buffer	= pinfo->buffer + offset;
	pinfo->size		= pinfo->size - offset;
	pinfo->realsize	= pinfo->realsize - offset;
	if(pinfo->size > udphdr.uh_ulen)
	{
		pinfo->size = udphdr.uh_ulen;
	}
	pinfo->src_port	= udphdr.uh_sport;
	pinfo->dst_port	= udphdr.uh_dport;
	pinfo->protocol = PROTOCOL_UDP;
	rc = dump_target(wdh, pinfo, err);
	return(rc);
}

int check_dump_target(wtap_dumper *wdh, const struct wtap_pkthdr* phdr, const unsigned char* pd, int *err)
{
	int			offset;
	int			size;
	int			realsize;
	int			rc;
	ether_hdr	etherhdr;
	ip_hdr		iphdr;
	packet_info	pinfo;

	//ethernett[
	if(phdr->caplen < sizeof(etherhdr))
	{
		return(ERROR_PACKET_TOO_SHORT);
	}
	etherhdr = *(ether_hdr*)pd;
	etherhdr.frametype = ntohs(etherhdr.frametype);
	if(etherhdr.frametype != FRAMETYPE_IP)
	{
		return(ERROR_PACKET_NOT_IP);
	}
	offset = sizeof(ether_hdr);
	//hopPbg
	if(phdr->caplen < offset + sizeof(ip_hdr))
	{
		return(ERROR_PACKET_TOO_SHORT);
	}
	iphdr = *(ip_hdr*)(pd + offset);
	iphdr.ip_len	= ntohs(iphdr.ip_len);
	iphdr.ip_id		= ntohs(iphdr.ip_id);
	iphdr.ip_off	= ntohs(iphdr.ip_off);
	iphdr.ip_sum	= ntohs(iphdr.ip_sum);
	iphdr.ip_src4	= ntohl(iphdr.ip_src4);
	iphdr.ip_dst4	= ntohl(iphdr.ip_dst4);
	rc = (iphdr.ip_v_hl & 0x0f) << 2;
	if(((iphdr.ip_v_hl & 0xf0) >> 4) != IPv4)
	{	//not IPv4
		return(ERROR_PACKET_IPv4);
	}
	if(rc < 20)
	{	//IP header size invalid
		return(ERROR_PACKET_IP_INVALID_SIZE);
	}
	if((iphdr.ip_len < (unsigned int)rc) || ((unsigned int)offset + iphdr.ip_len > phdr->len))
	{	//IP header size conflict
		return(ERROR_PACKET_SIZE_CONFLICT);
	}
	if((unsigned int)offset + iphdr.ip_len > phdr->caplen)
	{	//partial dump
		size = phdr->caplen - offset - rc;
	}
	else
	{	//all dump
		size = iphdr.ip_len - rc;
	}
	realsize = iphdr.ip_len - rc;
	offset += rc;
	if(wdh->fh->ulSourceIP != 0)
	{
		if(wdh->fh->ulSourceIP != iphdr.ip_src4)
		{
			return(ERROR_PACKET_NOT_TARGET);
		}
	}
	if(wdh->fh->ulDestinationIP != 0)
	{
		if(wdh->fh->ulDestinationIP != iphdr.ip_dst4)
		{
			return(ERROR_PACKET_NOT_TARGET);
		}
	}
	pinfo.src_ip	= iphdr.ip_src4;
	pinfo.dst_ip	= iphdr.ip_dst4;
	pinfo.buffer	= pd + offset;
	pinfo.size		= size;
	pinfo.realsize	= realsize;
	switch(iphdr.ip_p)
	{
	case 0x06:		//TCP
		if(wdh->fh->ulPortocol & PROTOCOL_TCP)
		{
			rc = checkTCP(wdh, &pinfo, err);
		}
		else
		{
			return(ERROR_PACKET_PROTOCOL_NOT_SUPPORT);
		}
		break;
// 2004/07/02 change start
//	case 11:	//UDP
	case 0x11:	//UDP
// 2004/07/02 change end
		if(wdh->fh->ulPortocol & PROTOCOL_UDP)
		{
			rc = checkUDP(wdh, &pinfo, err);
		}
		else
		{
			return(ERROR_PACKET_PROTOCOL_NOT_SUPPORT);
		}
		break;
	default:	//etc
		return(ERROR_PACKET_PROTOCOL_NOT_SUPPORT);
	}
	return(rc);
}
